/*
 * This file is part of the Picasso Visualization Environment
 *
 * Developed by: The TAPESTRY Parallel Computing Laboratory
 *		 University of Illinois at Urbana-Champaign
 *		 Department of Computer Science
 *		 1304 W. Springfield Avenue
 *		 Urbana, IL	61801
 *
 * Copyright (c) 1987, 1988, 1989, 1990
 * The University of Illinois Board of Trustees.
 *	All Rights Reserved.
 *
 * CONFIDENTIAL INFORMATION. Distribution
 * restricted under license agreement.
 *
 * Author: Allen D. Malony (malony@uicsrd.csrd.uiuc.edu)
 * Project Manager and Principal Investigator:
 *	Daniel A. Reed (reed@cs.uiuc.edu)
 *
 * Funded by: National Science Foundation grants NSF CCR86-57696,
 * NSF CCR87-06653 and NSF CDA87-22836 (Tapestry), NASA ICLASS Contract
 * No. NAG-1-613 and by grants from the Digital Equipment Corporation
 * External Research Program and Apple Computer.
 *
 */
/*
 *	$Header: /users/c/chaos/CVSmaster/bounce_demo/gui/Dial.c,v 1.2 1998/08/21 15:35:34 eisen Exp $
 */

#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/StringDefs.h>
#include <X11/IntrinsicP.h>
#include "../config.h"
#include "DialP.h"
#ifndef PRINTF_DEFINED
extern int printf(const char *, ...);
#endif

#define Min(a, b)	((a) < (b) ? (a) : (b))
#define Round(x)	(x >= 0.0 ? (int)(x + .5) : (int)(x - .5))
#define Offset(field)	XtOffset(DialWidget,field)
#define Norm(w, v)	(((float)v - (float)w->dial.min_value) /	\
			 ((float)w->dial.max_value -			\
			  (float)w->dial.min_value))

double	cos(), sin();

/**********************/
/* private procedures */
/**********************/
static void	ClassInitialize(), Initialize();
static void	Destroy(), Resize(), Redisplay();
static Boolean	SetValues();

static void	DestroyGCs(), CreateGCs();
static void	DrawNeedle(), DrawDialFace();
static void 	ComputeTickPoints(), ComputeNeedlePoints();

static void	DialSelect(), DialNotify();

/*********************/
/* public procedures */
/*********************/
extern void DialSetValue();

/*********************/
/* dial translations */
/*********************/
static char dial_translations[] = "	\
  <Btn1Down>:	DialSelect()\n		\
  <Btn1Up>:	DialNotify()		\
";

/****************/
/* dial actions */
/****************/
static XtActionsRec	dial_actions[] = {
  {"DialSelect",	(XtActionProc) DialSelect	},
  {"DialNotify",	(XtActionProc) DialNotify	},
};

/******************/
/* dial resources */
/******************/
static XtResource resources[] = {
  /******************/
  /* core resources */
  /******************/
/*
  {XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
     Offset(core.width), XtRString, "0"},
  {XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
     Offset(core.height), XtRString, "0"},
  {XtNbackground, XtCBackground, XtRPixel, sizeof(Pixel),
     Offset(core.background_pixel), XtRString, "white"},
*/
  /******************/
  /* dial resources */
  /******************/
  {XtNbeginDegree, XtCBeginDegree, XtRInt, sizeof(int),
     Offset(dial.begin_degree), XtRString,"0"},
  {XtNendDegree, XtCEndDegree, XtRInt, sizeof(int),
     Offset(dial.end_degree), XtRString,"360"},
  {XtNfaceBorderColor, XtCFaceBorderColor, XtRPixel, sizeof(Pixel),
     Offset(dial.face_border_color), XtRString, "black"},
  {XtNfaceColor, XtCFaceColor, XtRPixel, sizeof(Pixel),
     Offset(dial.face_color), XtRString, "white"},
  {XtNneedleColor, XtCNeedleColor, XtRPixel, sizeof(Pixel),
     Offset(dial.needle_color), XtRString, "black"},
  {XtNneedleBorderColor, XtCNeedleBorderColor, XtRPixel, sizeof(Pixel),
     Offset(dial.needle_border_color), XtRString, "black"},
  {XtNtickColor, XtCTickColor, XtRPixel, sizeof(Pixel),
     Offset(dial.tick_color), XtRString, "black"},
  {XtNminValue, XtCMinValue, XtRInt, sizeof(int),
     Offset(dial.min_value), XtRString, "0"},
  {XtNmaxValue, XtCMaxValue, XtRInt, sizeof(int),
     Offset(dial.max_value), XtRString, "100"},
  {XtNselect, XtCCallback, XtRCallback, sizeof(caddr_t),
     Offset(dial.select), XtRCallback, (caddr_t) NULL},
};
#undef Offset

/*********************/
/* dial class record */
/*********************/
DialClassRec dialClassRec = {
  &widgetClassRec,			/* superclass			*/
  "Dial",				/* class_name			*/
  sizeof(DialRec),			/* widget_size			*/
  ClassInitialize,			/* class_initialize		*/
  NULL,					/* class_part_initialize	*/
  FALSE,				/* class_inited			*/
  Initialize,				/* initialize			*/
  NULL,					/* initialize_hook		*/
  XtInheritRealize,			/* realize			*/
  dial_actions,				/* actions			*/
  XtNumber(dial_actions),		/* num_actions			*/
  resources,				/* resources			*/
  XtNumber(resources),			/* resource_count		*/
  NULL,					/* xrm_class			*/
  TRUE,					/* compress_motion		*/
  TRUE,					/* compress_exposure		*/
  TRUE,					/* compress_enterleave		*/
  FALSE,				/* visible_interest		*/
  Destroy,				/* destroy			*/
  Resize,				/* resize			*/
  Redisplay,				/* expose			*/
  SetValues,				/* set_values			*/
  NULL,					/* set_values_hook		*/
  XtInheritSetValuesAlmost,		/* set_values_almost		*/
  NULL,					/* get_values_hook		*/
  NULL,					/* accept_focus			*/
  XtVersion,				/* version			*/
  NULL,					/* callback_private		*/
  dial_translations,			/* tm_table			*/
  XtInheritQueryGeometry,		/* query_geometry		*/
  XtInheritDisplayAccelerator,		/* display_accelerator		*/
};

WidgetClass dialWidgetClass = (WidgetClass) &dialClassRec;

/************************************************************************/
/*                              PRIVATE PROCEDURES                      */
/************************************************************************/

/************************************************************************/
/* ClassInitialize()							*/
/************************************************************************/
static void ClassInitialize()
{
#ifdef DEBUG
printf("ClassInitialize\n");
#endif
  XtAddConverter(XtRString, XtRBackingStore, XmuCvtStringToBackingStore,
		 NULL, 0);
}

/* ARGSUSED */

/************************************************************************/
/* Initialize()								*/
/************************************************************************/
static void Initialize (request, new)
Widget request, new;
{
  DialWidget	w = (DialWidget)new;
  int		min_height, min_width;

#ifdef DEBUG
printf("Initialize\n");
#endif
  min_width = min_height = MIN_SIZE;
  if (w->core.width == 0)
    w->core.width = min_width;
  if (w->core.height == 0)
    w->core.height = min_height;
  
  CreateGCs(w);

  w->dial.needle_length = 0;
  w->dial.needle_width = 0;
  w->dial.centerX = 0;
  w->dial.centerY = 0;
  w->dial.value = w->dial.min_value;
  Resize(w);
  ComputeNeedlePoints(w);

} /* Initialize */

/************************************************************************/
/* SetValues()								*/
/************************************************************************/
static Boolean SetValues(gcurrent, grequest, gnew)
Widget	gcurrent, grequest, gnew;
{
  DialWidget	current = (DialWidget) gcurrent;
  DialWidget	new = (DialWidget) gnew;
  Boolean	redisplay = FALSE;
  XtGCMask	GCmask;
  XGCValues	GCvalues;

#ifdef DEBUG
printf("SetValues\n");
#endif
  if (new->dial.face_border_color != current->dial.face_border_color) {
    XtDestroyGC(current->dial.face_borderGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->dial.face_border_color;
    new->dial.face_borderGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }
  
  if (new->dial.face_color != current->dial.face_color) {
    XtDestroyGC(current->dial.faceGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->dial.face_color;
    new->dial.faceGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }
  
  if (new->core.background_pixel != current->core.background_pixel) {
    XtDestroyGC(current->dial.eraseGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->core.background_pixel;
    new->dial.eraseGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }

  if (new->dial.needle_color != current->dial.needle_color) {
    XtDestroyGC(current->dial.needleGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->dial.needle_color;
    new->dial.needleGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }

  if (new->dial.needle_border_color != current->dial.needle_border_color) {
    XtDestroyGC(current->dial.needle_borderGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->dial.needle_border_color;
    new->dial.needle_borderGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }

  if (new->dial.needle_color != current->dial.needle_color) {
    XtDestroyGC(current->dial.needleGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->dial.needle_color;
    new->dial.needleGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }

  if (new->dial.tick_color != current->dial.tick_color) {
    XtDestroyGC(current->dial.tickGC);
    GCmask = GCForeground;
    GCvalues.foreground = new->dial.tick_color;
    new->dial.tickGC = XtGetGC((Widget) new, GCmask, &GCvalues);
    redisplay = TRUE;
  }

  return(redisplay);
  
} /* SetValues */

/************************************************************************/
/* Destroy()								*/
/************************************************************************/
static void Destroy (gw)
Widget gw;
{
  DialWidget	w = (DialWidget) gw;

#ifdef DEBUG
printf("Destroy\n");
#endif
  DestroyGCs(w);
  XtRemoveAllCallbacks(gw, XtNselect);

} /* Destroy */

/************************************************************************/
/* Resize()								*/
/************************************************************************/
static void Resize (gw) 
Widget gw;
{
  DialWidget	w = (DialWidget) gw;
  int		height;

#ifdef DEBUG
printf("Resize\n");
#endif
/*
  if (XtIsRealized(gw)) {
*/
    height = w->core.height;
    w->dial.radius = ((int) Min(w->core.width, height) - 
		      (int) (2 * PADDING)) / 2;
    w->dial.needle_length = NEEDLE_FRACTION * w->dial.radius;
    w->dial.needle_width = NEEDLE_WIDTH_FRACTION * w->dial.radius;
    w->dial.centerX = w->core.width / 2;
    w->dial.centerY = height / 2;
/*
  }
*/
  ComputeNeedlePoints(w);

} /* Resize */

/************************************************************************/
/* Redisplay()								*/
/************************************************************************/
static void Redisplay(gw, event, region)
Widget	gw;
XEvent	*event;
Region	region;
{
  DialWidget	w = (DialWidget) gw;
  
#ifdef DEBUG
printf("Redisplay\n");
#endif
  XClearArea(XtDisplay(w), XtWindow(w),
	     0, 0, w->core.width, w->core.height, False);
  DrawDialFace(w);
  DrawNeedle(w);

} /* Redisplay */

/************************************************************************/
/* DestroyGCs()								*/
/************************************************************************/
static void DestroyGCs(w)
DialWidget	w;
{

#ifdef DEBUG
printf("DestroyGCs\n");
#endif
  XtDestroyGC(w->dial.face_borderGC);
  XtDestroyGC(w->dial.faceGC);
  XtDestroyGC(w->dial.eraseGC);
  XtDestroyGC(w->dial.needleGC);
  XtDestroyGC(w->dial.needle_borderGC);
  XtDestroyGC(w->dial.tickGC);

} /* DestroyGCs */

/************************************************************************/
/* CreateGCs()								*/
/************************************************************************/
static void CreateGCs(w)
DialWidget	w;
{
  XtGCMask	GCmask;
  XGCValues	GCvalues;

#ifdef DEBUG
printf("CreateGCs\n");
#endif
  /* create face border GC */
  GCmask = GCForeground;
  GCvalues.foreground = w->dial.face_border_color;
  w->dial.face_borderGC = XtGetGC((Widget)w, GCmask, &GCvalues);
  
  /* create face GC */
  GCmask = GCForeground;
  GCvalues.foreground = w->dial.face_color;
  w->dial.faceGC = XtGetGC((Widget)w, GCmask, &GCvalues);
  
  /* creat erase GC */
  GCmask = GCForeground;
  GCvalues.foreground = w->core.background_pixel;
  w->dial.eraseGC = XtGetGC((Widget)w, GCmask, &GCvalues);

  /* create needle GC */
  GCmask = GCForeground;
  GCvalues.foreground = w->dial.needle_color;
  w->dial.needleGC = XtGetGC((Widget)w, GCmask, &GCvalues);
  
  /* create needle border GC */
  GCmask = GCForeground;
  GCvalues.foreground = w->dial.needle_border_color;
  w->dial.needle_borderGC = XtGetGC((Widget)w, GCmask, &GCvalues);
  
  /* create tick GC */
  GCmask = GCForeground;
  GCvalues.foreground = w->dial.tick_color;
  w->dial.tickGC = XtGetGC((Widget)w, GCmask, &GCvalues);

} /* CreateGCs */

/************************************************************************/
/* EraseNeedle()							*/
/************************************************************************/
static void EraseNeedle(w)
DialWidget	w;
{
#ifdef DEBUG
printf("EraseNeedle\n");
#endif
  if (!XtIsRealized((Widget)w)) return;
  XDrawLines(XtDisplay(w), XtWindow(w), w->dial.faceGC,
	     w->dial.needle, NEEDLE_POINTS, CoordModeOrigin);
  XFillPolygon(XtDisplay(w), XtWindow(w), w->dial.faceGC,
	       w->dial.needle, NEEDLE_POINTS,
	       Convex, CoordModeOrigin);

} /* EraseNeedle */

/************************************************************************/
/* DrawNeedle()								*/
/************************************************************************/
static void DrawNeedle(w)
DialWidget	w;
{
#ifdef DEBUG
printf("DrawNeedle\n");
#endif
  if (!XtIsRealized((Widget)w)) return;
  XFillPolygon(XtDisplay(w), XtWindow(w), w->dial.needleGC,
	       w->dial.needle, NEEDLE_POINTS,
	       Convex, CoordModeOrigin);
  XDrawLines(XtDisplay(w), XtWindow(w), w->dial.needle_borderGC,
	     w->dial.needle, NEEDLE_POINTS, CoordModeOrigin);

} /* DrawNeedle */

/************************************************************************/
/* ComputeNeedlePoints()						*/
/************************************************************************/
static void ComputeNeedlePoints(w)
DialWidget	w;
{
  Dimension		length = w->dial.needle_length;
  Dimension		width = w->dial.needle_width;
  register double	angle, cosangle, sinangle;
  register double	ws, wc;

#ifdef DEBUG
printf("ComputeNeedlePoints\n");
#endif
  /**********************************************************************/
  /* A full circle is 2 PI radians.  Angles are measured from 12	*/
  /* o'dial, dialwise increasing.					*/
  /**********************************************************************/
  angle = PIHALF - TWOPI * (double) Norm(w, w->dial.value);
  cosangle = cos(angle);
  sinangle = sin(angle);

  /**********************************************************************/
  /* Order of points of the needle.					*/
  /*									*/
  /*		0,3							*/
  /*		/ \							*/
  /*	       /   \							*/
  /*	      /     \							*/
  /*	     1-------2							*/
  /**********************************************************************/
  wc = Round(width * cosangle);
  ws = Round(width * sinangle);

  /* 0 */
  w->dial.needle[0].x = w->dial.centerX + Round(length * cosangle);
  w->dial.needle[0].y = w->dial.centerY - Round(length * sinangle);
  /* 1 */
  w->dial.needle[1].x = w->dial.centerX - Round(wc - ws); 
  w->dial.needle[1].y = w->dial.centerY + Round(ws + wc);
  /* 2 */
  w->dial.needle[2].x = w->dial.centerX - Round(wc + ws);
  w->dial.needle[2].y = w->dial.centerY + Round(ws - wc);
  /* 3 */
  w->dial.needle[3].x = w->dial.centerX + Round(length * cosangle);
  w->dial.needle[3].y = w->dial.centerY - Round(length * sinangle);

} /* ComputeNeedlePoints */

/************************************************************************/
/* ComputeTickPoints()							*/
/************************************************************************/
static void ComputeTickPoints(w, length, fraction, s)
DialWidget	w;
Dimension	length;
double		fraction;
XSegment	*s;
{
  int		cx = w->dial.centerX;
  int		cy = w->dial.centerY;
  double	angle, cosangle, sinangle;

#ifdef DEBUG
printf("ComputeTickPoints\n");
#endif
  /**********************************************************************/
  /* A full circle is 2 PI radians.  Angles are measured from 12	*/
  /* o'dial, dialwise increasing.					*/
  /**********************************************************************/
  angle = PIHALF - TWOPI * fraction;
  cosangle = cos(angle);
  sinangle = sin(angle);

  s->x1 = cx + (int)(length * sinangle);
  s->y1 = cy - (int)(length * cosangle);
  s->x2 = cx + (int)(w->dial.radius * sinangle);
  s->y2 = cy - (int)(w->dial.radius * cosangle);

} /* ComputeTickPoints */

/************************************************************************/
/* DrawDialFace()							*/
/************************************************************************/
static void DrawDialFace(w)
DialWidget	w;
{
  register int	i, s;
  Dimension	length = w->dial.needle_length;
  Dimension	delta = (w->dial.radius - length) / 2;

#ifdef DEBUG
printf("DrawDialFace\n");
#endif
  if (!XtIsRealized((Widget)w)) return;
  XFillArc(XtDisplay(w), XtWindow(w), w->dial.faceGC,
	   w->dial.centerX - w->dial.radius,
	   w->dial.centerY - w->dial.radius,
	   2*w->dial.radius, 2*w->dial.radius,
	   0, 360*64);
  s = 0;
  for (i=0; i<MAXTICKS; i++) {
    if ((i % MAJOR_TICK) == 0)
      ComputeTickPoints(w, length, ((double) i)/(double) MAXTICKS,
			&w->dial.ticks[s]);
    else
      ComputeTickPoints(w, length+delta, ((double) i)/(double) MAXTICKS,
			&w->dial.ticks[s]);
    s++;
  }

  /* draw the ticks */
  XDrawSegments(XtDisplay(w), XtWindow(w), w->dial.tickGC,
		w->dial.ticks, s);
	
  /* draw dial edge */
  /***** need to add filling of dial face with face color *****/
  XDrawArc(XtDisplay(w), XtWindow(w), w->dial.face_borderGC,
	   w->dial.centerX - w->dial.radius,
	   w->dial.centerY - w->dial.radius,
	   2*w->dial.radius, 2*w->dial.radius,
	   0, 360*64);

} /* DrawDialFace */

/************************************************************************/
/* DialSelect() processes button down event.				*/
/************************************************************************/
static void DialSelect(w, event)
DialWidget		w;
XButtonPressedEvent	*event;
{
#ifdef DEBUG
printf("DialSelect\n");
#endif
  w->dial.value_select = w->dial.value;

} /* DialSelect */

/************************************************************************/
/* DialNotify() processes button up event.				*/
/************************************************************************/
static void DialNotify(w, event)
DialWidget		w;
XButtonReleasedEvent	*event;
{
#ifdef DEBUG
printf("DialNotify\n");
#endif
  XtCallCallbacks((Widget)w, XtNselect, (caddr_t) event);

} /* DialNotify */

/************************************************************************/
/*                          PUBLIC PROCEDURES                           */
/************************************************************************/

/************************************************************************/
/* DialSetValue() set the current dial value to v.			*/
/************************************************************************/
extern void DialSetValue(w, v)
DialWidget	w;
int		v;
{
#ifdef DEBUG
printf("DialSetValue\n");
#endif
  EraseNeedle(w);
  if (v < w->dial.min_value) {
    printf("DialSetValue: tried to set a value (%d) ", v);
    printf("less than min (%d)\n", w->dial.min_value);
    w->dial.value = w->dial.min_value;
  }
  else if (v > w->dial.max_value) { 
    printf("DialSetValue: tried to set a value (%d) ", v);
    printf("greater than max (%d)\n", w->dial.max_value);
    w->dial.value = w->dial.max_value;
  }
  else
    w->dial.value = v;
  ComputeNeedlePoints(w);
  if (XtIsRealized((Widget)w)) {
      DrawNeedle(w);
  }
} /* DialSetValue */
